@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2014-2015 Phusion Holding B.V.
+ *  Copyright (c) 2014-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -34,8 +34,8 @@
 #include <ServerKit/http_parser.h>
 #include <ServerKit/CookieUtils.h>
 #include <StaticString.h>
-#include <Utils/DateParsing.h>
-#include <Utils/StrIntUtils.h>
+#include <StrIntTools/DateParsing.h>
+#include <StrIntTools/StrIntUtils.h>
 
 namespace Passenger {
 
@@ -492,12 +492,12 @@ public:
 		}
 
 		LString *varyCookieName = req->secureHeaders.lookup(PASSENGER_VARY_TURBOCACHE_BY_COOKIE);
-		if (varyCookieName == NULL && !controller->defaultVaryTurbocacheByCookie.empty()) {
+		if (varyCookieName == NULL && !req->config->defaultVaryTurbocacheByCookie.empty()) {
 			varyCookieName = (LString *) psg_palloc(req->pool, sizeof(LString));
 			psg_lstr_init(varyCookieName);
 			psg_lstr_append(varyCookieName, req->pool,
-				controller->defaultVaryTurbocacheByCookie.data(),
-				controller->defaultVaryTurbocacheByCookie.size());
+				req->config->defaultVaryTurbocacheByCookie.data(),
+				req->config->defaultVaryTurbocacheByCookie.size());
 		}
 		if (varyCookieName != NULL) {
 			LString *cookieHeader = req->headers.lookup(COOKIE);
