@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010-2016 Phusion Holding B.V.
+ *  Copyright (c) 2010-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -38,30 +38,39 @@
  * prepareRequest() and handleRequest() can be comfortably run.
  */
 
+// In Apache < 2.4, this macro was necessary for core_dir_config and other structs
+#define CORE_PRIVATE
+
 #include <boost/thread.hpp>
 
 #include <sys/time.h>
 #include <sys/resource.h>
+#include <sys/stat.h>
 #include <exception>
 #include <cstdio>
+#include <fcntl.h>
 #include <unistd.h>
 
 #include <oxt/initialize.hpp>
 #include <oxt/macros.hpp>
 #include <oxt/backtrace.hpp>
 #include <oxt/detail/context.hpp>
-#include "Hooks.h"
 #include "Bucket.h"
-#include "Configuration.hpp"
+#include "Config.h"
 #include "DirectoryMapper.h"
+#include "Utils.h"
 #include <modp_b64.h>
+#include <WrapperRegistry/Registry.h>
+#include <FileTools/FileManip.h>
+#include <FileTools/FileManip.h>
 #include <Utils.h>
-#include <Utils/IOUtils.h>
-#include <Utils/StrIntUtils.h>
-#include <Utils/SystemTime.h>
+#include <IOTools/IOUtils.h>
+#include <StrIntTools/StrIntUtils.h>
+#include <SystemTools/SystemTime.h>
 #include <Utils/HttpConstants.h>
 #include <Utils/ReleaseableScopedPointer.h>
-#include <Logging.h>
+#include <LoggingKit/LoggingKit.h>
+#include <LoggingKit/Context.h>
 #include <WatchdogLauncher.h>
 #include <Constants.h>
 
@@ -82,8 +91,8 @@
 #include <apr_lib.h>
 #include <unixd.h>
 
-using namespace std;
-using namespace Passenger;
+#include "DirConfig/AutoGeneratedHeaderSerialization.cpp"
+
 
 extern "C" module AP_MODULE_DECLARE_DATA passenger_module;
 
@@ -102,11 +111,12 @@ extern "C" module AP_MODULE_DECLARE_DATA passenger_module;
 #endif
 
 
-/**
- * Apache hook functions, wrapped in a class.
- *
- * @ingroup Core
- */
+namespace Passenger {
+namespace Apache2Module {
+
+using namespace std;
+
+
 class Hooks {
 private:
 	class ErrorReport {
@@ -223,9 +233,36 @@ private:
 	enum Threeway { YES, NO, UNKNOWN };
 
 	Threeway m_hasModRewrite, m_hasModDir, m_hasModAutoIndex, m_hasModXsendfile;
+	WrapperRegistry::Registry wrapperRegistry;
 	CachedFileStat cstat;
 	WatchdogLauncher watchdogLauncher;
 	boost::mutex cstatMutex;
+	boost::mutex configMutex;
+
+	static Json::Value strsetToJson(const set<string> &input) {
+		Json::Value result(Json::arrayValue);
+		set<string>::const_iterator it, end = input.end();
+		for (it = input.begin(); it != end; it++) {
+			result.append(*it);
+		}
+		return result;
+	}
+
+	static Json::Value nonEmptyString(const char *str) {
+		if (str != NULL && *str != '\0') {
+			return str;
+		} else {
+			return Json::Value();
+		}
+	}
+
+	static Json::Value nonEmptyString(const string &str) {
+		if (str.empty()) {
+			return Json::Value();
+		} else {
+			return str;
+		}
+	}
 
 	inline DirConfig *getDirConfig(request_rec *r) {
 		return (DirConfig *) ap_get_module_config(r->per_dir_config, &passenger_module);
@@ -358,6 +395,27 @@ private:
 		return m_hasModXsendfile == YES;
 	}
 
+	static bool stderrEqualsFile(const char *path) {
+		struct stat s1, s2;
+
+		if (fstat(STDERR_FILENO, &s1) == -1) {
+			return false;
+		}
+
+		// No O_CREAT: we don't care if the file does not exist.
+		int fd = open(path, O_WRONLY | O_APPEND, 0600);
+		if (fd == -1) {
+			return false;
+		}
+		if (fstat(fd, &s2) == -1) {
+			close(fd);
+			return false;
+		}
+		close(fd);
+
+		return s1.st_dev == s2.st_dev && s1.st_ino == s2.st_ino && s1.st_rdev == s2.st_rdev;
+	}
+
 	int reportBusyException(request_rec *r) {
 		ap_custom_response(r, HTTP_SERVICE_UNAVAILABLE,
 			"This website is too busy right now.  Please try again later.");
@@ -379,7 +437,7 @@ private:
 	 *   handler shouldn't be run.
 	 * - If D is not true, then the handler should be run.
 	 *
-	 * @pre config->isEnabled()
+	 * @pre config->getEnabled()
 	 * @param coreModuleWillBeRun Whether the core.c map_to_storage hook might be called after this.
 	 * @return Whether the Phusion Passenger handler hook method should be run.
 	 *         When true, this method will save a request note object so that future hooks
@@ -388,9 +446,12 @@ private:
 	bool prepareRequest(request_rec *r, DirConfig *config, const char *filename, bool coreModuleWillBeRun = false) {
 		TRACE_POINT();
 
-		DirectoryMapper mapper(r, config, &cstat, &cstatMutex, serverConfig.statThrottleRate);
+		DirectoryMapper mapper(r, config, wrapperRegistry, &cstat,
+		    &cstatMutex, serverConfig.statThrottleRate, &configMutex);
 		try {
-			if (mapper.getApplicationType() == PAT_NONE) {
+			if (config->getAppStartCommand().empty()
+			 && mapper.getDetectorResult().isNull())
+			{
 				// (B) is not true.
 				disableRequestNote(r);
 				return false;
@@ -567,8 +628,8 @@ private:
 				return httpStatus;
 			}
 
-			this_thread::disable_interruption di;
-			this_thread::disable_syscall_interruption dsi;
+			boost::this_thread::disable_interruption di;
+			boost::this_thread::disable_syscall_interruption dsi;
 			bool expectingBody;
 
 			expectingBody = ap_should_client_block(r);
@@ -657,7 +718,7 @@ private:
 				apr_table_setn(r->headers_out, "Status", r->status_line);
 
 				UPDATE_TRACE_POINT();
-				if (config->errorOverride == DirConfig::ENABLED
+				if (config->getErrorOverride()
 				 && ap_is_HTTP_ERROR(r->status))
 				{
 					/* Send ErrorDocument.
@@ -801,44 +862,6 @@ private:
 		return strstr(buffer, "upgrade");
 	}
 
-	void addHeader(string &headers, const StaticString &name, const char *value) {
-		if (value != NULL) {
-			headers.append(name.data(), name.size());
-			headers.append(": ", 2);
-			headers.append(value);
-			headers.append("\r\n", 2);
-		}
-	}
-
-	void addHeader(string &headers, const StaticString &name, const StaticString &value) {
-		headers.append(name.data(), name.size());
-		headers.append(": ", 2);
-		headers.append(value.data(), value.size());
-		headers.append("\r\n", 2);
-	}
-
-	void addHeader(request_rec *r, string &headers, const StaticString &name, int value) {
-		if (value != UNSET_INT_VALUE) {
-			headers.append(name.data(), name.size());
-			headers.append(": ", 2);
-			headers.append(apr_psprintf(r->pool, "%d", value));
-			headers.append("\r\n", 2);
-		}
-	}
-
-	void addHeader(string &headers, const StaticString &name, DirConfig::Threeway value) {
-		if (value != DirConfig::UNSET) {
-			headers.append(name.data(), name.size());
-			headers.append(": ", 2);
-			if (value == DirConfig::ENABLED) {
-				headers.append("t", 1);
-			} else {
-				headers.append("f", 1);
-			}
-			headers.append("\r\n", 2);
-		}
-	}
-
 	string constructRequestHeaders(request_rec *r, DirectoryMapper &mapper,
 		bool &bodyIsChunked)
 	{
@@ -852,7 +875,7 @@ private:
 		result.append(r->method);
 		result.append(" ", 1);
 
-		if (config->allowsEncodedSlashes()) {
+		if (config->getAllowEncodedSlashes()) {
 			/*
 			 * Apache decodes encoded slashes in r->uri, so we must use r->unparsed_uri
 			 * if we are to support encoded slashes. However mod_rewrite doesn't change
@@ -952,30 +975,33 @@ private:
 		addHeader(result, P_STATIC_STRING("!~REMOTE_USER"), r->user);
 
 		// App group name.
-		if (config->appGroupName == NULL) {
+		if (config->getAppGroupName().empty()) {
 			result.append("!~PASSENGER_APP_GROUP_NAME: ",
 				sizeof("!~PASSENGER_APP_GROUP_NAME: ") - 1);
 			result.append(mapper.getAppRoot());
-			if (config->appEnv != NULL) {
+			if (!config->getAppEnv().empty()) {
 				result.append(" (", 2);
-				result.append(config->appEnv);
+				result.append(config->getAppEnv().data(),
+					config->getAppEnv().size());
 				result.append(")", 1);
 			}
 			result.append("\r\n", 2);
 		}
 
 		// Phusion Passenger options.
-		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_ROOT"), mapper.getAppRoot());
-		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_TYPE"), mapper.getApplicationTypeName());
-		if (config->useUnionStation() && !config->unionStationKey.empty()) {
-			addHeader(result, P_STATIC_STRING("!~UNION_STATION_SUPPORT"), P_STATIC_STRING("t"));
-			addHeader(result, P_STATIC_STRING("!~UNION_STATION_KEY"), config->unionStationKey);
-			if (!config->unionStationFilters.empty()) {
-				addHeader(result, P_STATIC_STRING("!~UNION_STATION_FILTERS"),
-					config->getUnionStationFilterString());
-			}
+		addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_ROOT"),
+			mapper.getAppRoot());
+		if (!config->getAppStartCommand().empty()) {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_START_COMMAND"),
+				config->getAppStartCommand());
+		} else if (mapper.getDetectorResult().wrapperRegistryEntry != NULL) {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_TYPE"),
+				mapper.getDetectorResult().wrapperRegistryEntry->language);
+		} else {
+			addHeader(result, P_STATIC_STRING("!~PASSENGER_APP_START_COMMAND"),
+				mapper.getDetectorResult().appStartCommand);
 		}
-		#include "SetHeaders.cpp"
+		constructRequestHeaders_autoGenerated(r, config, result);
 
 		/*********************/
 		/*********************/
@@ -994,6 +1020,10 @@ private:
 			env = (apr_table_entry_t*) env_arr->elts;
 
 			for (i = 0; i < env_arr->nelts; ++i) {
+				if ((strcmp(env[i].key, "SCRIPT_NAME") == 0)
+				 || (strcmp(env[i].key, "PATH_INFO")   == 0)) {
+					continue;
+				}
 				envvarsData.append(env[i].key);
 				envvarsData.append("\0", 1);
 				if (env[i].val != NULL) {
@@ -1028,7 +1058,7 @@ private:
 		// S = SSL
 
 		result.append("!~FLAGS: CD", sizeof("!~FLAGS: CD") - 1);
-		if (config->bufferUpload != DirConfig::DISABLED) {
+		if (config->getBufferUpload()) {
 			result.append("B", 1);
 		}
 		if (lookupEnv(r, "HTTPS") != NULL) {
@@ -1237,22 +1267,37 @@ public:
 	    : cstat(1024),
 	      watchdogLauncher(IM_APACHE)
 	{
-		passenger_postprocess_config(s);
+		wrapperRegistry.finalize();
+		postprocessConfig(s, pconf, ptemp);
 
-		Passenger::setLogLevel(serverConfig.logLevel);
-		if (serverConfig.logFile != NULL) {
-			int errcode;
-			if (!Passenger::setLogFileWithoutRedirectingStderr(serverConfig.logFile, &errcode)) {
-				fprintf(stderr,
-					"ERROR: cannot open log file %s: %s (errno=%d)\n",
-					serverConfig.logFile,
-					strerror(errcode),
-					errcode);
-			}
+		Json::Value loggingConfig;
+		loggingConfig["level"] = LoggingKit::Level(serverConfig.logLevel);
+		loggingConfig["redirect_stderr"] = false;
+		if (!serverConfig.logFile.empty()) {
+			loggingConfig["target"] = serverConfig.logFile.toString();
 		}
-		if (serverConfig.fileDescriptorLogFile != NULL) {
-			Passenger::setFileDescriptorLogFile(serverConfig.fileDescriptorLogFile);
+		if (!serverConfig.fileDescriptorLogFile.empty()) {
+			loggingConfig["file_descriptor_log_target"] =
+				serverConfig.fileDescriptorLogFile.toString();
 		}
+
+		vector<ConfigKit::Error> errors;
+		LoggingKit::ConfigChangeRequest req;
+		bool ok;
+		try {
+			ok = LoggingKit::context->prepareConfigChange(loggingConfig,
+				errors, req);
+		} catch (const std::exception &e) {
+			ok = false;
+			fprintf(stderr, "ERROR: unable to configure logging system: %s\n", e.what());
+		}
+		if (ok) {
+			LoggingKit::context->commitConfigChange(req);
+		} else {
+			fprintf(stderr, "ERROR: unable to configuring logging system: %s\n",
+				ConfigKit::toString(errors).c_str());
+		}
+
 		m_hasModRewrite = UNKNOWN;
 		m_hasModDir = UNKNOWN;
 		m_hasModAutoIndex = UNKNOWN;
@@ -1261,7 +1306,7 @@ public:
 		P_DEBUG("Initializing Phusion Passenger...");
 		ap_add_version_component(pconf, SERVER_TOKEN_NAME "/" PASSENGER_VERSION);
 
-		if (serverConfig.root == NULL) {
+		if (serverConfig.root.empty()) {
 			throw ConfigurationException("The 'PassengerRoot' configuration option "
 				"is not specified. This option is required, so please specify it. "
 				"TIP: The correct value for this option was given to you by "
@@ -1274,37 +1319,50 @@ public:
 			const char *webServerDesc = ap_get_server_version();
 		#endif
 
-		VariantMap params;
-		params
-			.setPid ("web_server_control_process_pid", getpid())
-			.set    ("server_software", webServerDesc)
-			.setBool("multi_app", true)
-			.setBool("load_shell_envvars", true)
-			.set    ("file_descriptor_log_file", (serverConfig.fileDescriptorLogFile == NULL)
-				? "" : serverConfig.fileDescriptorLogFile)
-			.setInt	("socket_backlog", serverConfig.socketBacklog)
-			.set    ("data_buffer_dir", serverConfig.dataBufferDir)
-			.set    ("instance_registry_dir", serverConfig.instanceRegistryDir)
-			.setBool("user_switching", serverConfig.userSwitching)
-			.set    ("default_user", serverConfig.defaultUser)
-			.set    ("default_group", serverConfig.defaultGroup)
-			.set    ("default_ruby", serverConfig.defaultRuby)
-			.setInt ("max_pool_size", serverConfig.maxPoolSize)
-			.setInt ("pool_idle_time", serverConfig.poolIdleTime)
-			.setInt ("response_buffer_high_watermark", serverConfig.responseBufferHighWatermark)
-			.setInt ("stat_throttle_rate", serverConfig.statThrottleRate)
-			.set    ("analytics_log_user", serverConfig.analyticsLogUser)
-			.set    ("analytics_log_group", serverConfig.analyticsLogGroup)
-			.setBool("union_station_support", serverConfig.unionStationSupport)
-			.set    ("union_station_gateway_address", serverConfig.unionStationGatewayAddress)
-			.setInt ("union_station_gateway_port", serverConfig.unionStationGatewayPort)
-			.set    ("union_station_gateway_cert", serverConfig.unionStationGatewayCert)
-			.set    ("union_station_proxy_address", serverConfig.unionStationProxyAddress)
-			.setBool("turbocaching", serverConfig.turbocaching)
-			.setStrSet("prestart_urls", serverConfig.prestartURLs);
+		ap_version_t version;
+		ap_get_server_revision(&version);
+		string webServerVersion = toString(version.major) + "." + toString(version.minor) + "." + toString(version.patch);
+		if (version.add_string != NULL) {
+			webServerVersion.append(version.add_string);
+		}
 
-		if (serverConfig.logFile != NULL) {
-			params.set("log_file", serverConfig.logFile);
+		// Note: WatchdogLauncher::start() sets a number of default values.
+		Json::Value config;
+		config["web_server_module_version"] = PASSENGER_VERSION;
+		config["web_server_version"] = webServerVersion;
+		config["server_software"] = webServerDesc;
+		config["multi_app"] = true;
+		config["default_load_shell_envvars"] = true;
+		config["config_manifest"] = serverConfig.manifest;
+		config["file_descriptor_log_target"] = nonEmptyString(serverConfig.fileDescriptorLogFile);
+		config["controller_socket_backlog"] = serverConfig.socketBacklog;
+		config["controller_file_buffered_channel_buffer_dir"] = nonEmptyString(serverConfig.dataBufferDir);
+		config["instance_registry_dir"] = nonEmptyString(serverConfig.instanceRegistryDir);
+		config["spawn_dir"] = nonEmptyString(serverConfig.spawnDir);
+		config["security_update_checker_disabled"] = serverConfig.disableSecurityUpdateCheck;
+		config["security_update_checker_proxy_url"] = nonEmptyString(serverConfig.securityUpdateCheckProxy);
+		config["telemetry_collector_disabled"] = serverConfig.disableAnonymousTelemetry;
+		config["telemetry_collector_proxy_url"] = nonEmptyString(serverConfig.anonymousTelemetryProxy);
+		config["user_switching"] = serverConfig.userSwitching;
+		config["default_user"] = serverConfig.defaultUser.toString();
+		config["default_group"] = serverConfig.defaultGroup.toString();
+		config["default_ruby"] = serverConfig.defaultRuby.toString();
+		config["show_version_in_header"] = serverConfig.showVersionInHeader;
+		config["max_pool_size"] = serverConfig.maxPoolSize;
+		config["pool_idle_time"] = serverConfig.poolIdleTime;
+		config["max_instances_per_app"] = serverConfig.maxInstancesPerApp;
+		config["response_buffer_high_watermark"] = serverConfig.responseBufferHighWatermark;
+		config["stat_throttle_rate"] = serverConfig.statThrottleRate;
+		config["turbocaching"] = serverConfig.turbocaching;
+		config["prestart_urls"] = strsetToJson(serverConfig.prestartURLs);
+		config["admin_panel_url"] = nonEmptyString(serverConfig.adminPanelUrl);
+		config["admin_panel_auth_type"] = nonEmptyString(serverConfig.adminPanelAuthType);
+		config["admin_panel_username"] = nonEmptyString(serverConfig.adminPanelUsername);
+		config["admin_panel_password"] = nonEmptyString(serverConfig.adminPanelPassword);
+		config["disable_log_prefix"] = serverConfig.disableLogPrefix;
+
+		if (!serverConfig.logFile.empty()) {
+			config["log_target"] = serverConfig.logFile.toString();
 		} else if (s->error_fname == NULL) {
 			throw ConfigurationException("Cannot initialize " PROGRAM_NAME
 				" because Apache is not configured with an error log file."
@@ -1322,12 +1380,18 @@ public:
 				" support logging to syslog. Please configure " SHORT_PROGRAM_NAME
 				" with an explicit log file using the `PassengerLogFile` directive.");
 		} else {
-			params.set("log_file", ap_server_root_relative(pconf, s->error_fname));
+			config["log_target"]["path"] = ap_server_root_relative(pconf, s->error_fname);
+			if (stderrEqualsFile(ap_server_root_relative(pconf, s->error_fname))) {
+				config["log_target"]["stderr"] = true;
+			}
 		}
 
-		serverConfig.ctl.addTo(params);
+		Json::Value::iterator it, end = serverConfig.ctl.end();
+		for (it = serverConfig.ctl.begin(); it != end; it++) {
+			config[it.name()] = *it;
+		}
 
-		watchdogLauncher.start(serverConfig.root, params);
+		watchdogLauncher.start(serverConfig.root, config);
 	}
 
 	void childInit(apr_pool_t *pchild, server_rec *s) {
@@ -1336,7 +1400,9 @@ public:
 
 	int prepareRequestWhenInHighPerformanceMode(request_rec *r) {
 		DirConfig *config = getDirConfig(r);
-		if (config->isEnabled() && config->highPerformanceMode()) {
+		if (config->getEnabled()
+			&& config->getHighPerformance())
+		{
 			if (prepareRequest(r, config, r->filename, true)) {
 				return OK;
 			} else {
@@ -1377,8 +1443,8 @@ public:
 
 	int prepareRequestWhenNotInHighPerformanceMode(request_rec *r) {
 		DirConfig *config = getDirConfig(r);
-		if (config->isEnabled()) {
-			if (config->highPerformanceMode()) {
+		if (config->getEnabled()) {
+			if (config->getHighPerformance()) {
 				/* Preparations have already been done in the map_to_storage hook.
 				 * Prevent other modules' fixups hooks from being run.
 				 */
@@ -1517,7 +1583,7 @@ public:
 
 	int handleRequestWhenInHighPerformanceMode(request_rec *r) {
 		DirConfig *config = getDirConfig(r);
-		if (config->highPerformanceMode()) {
+		if (config->getHighPerformance()) {
 			return handleRequest(r);
 		} else {
 			return DECLINED;
@@ -1526,7 +1592,7 @@ public:
 
 	int handleRequestWhenNotInHighPerformanceMode(request_rec *r) {
 		DirConfig *config = getDirConfig(r);
-		if (config->highPerformanceMode()) {
+		if (config->getHighPerformance()) {
 			return DECLINED;
 		} else {
 			return handleRequest(r);
@@ -1540,19 +1606,16 @@ public:
  * Below follows lightweight C wrappers around the C++ Hook class.
  ******************************************************************/
 
-/**
- * @ingroup Hooks
- * @{
- */
-
 static Hooks *hooks = NULL;
 
 static apr_status_t
 destroy_hooks(void *arg) {
 	try {
-		this_thread::disable_interruption di;
-		this_thread::disable_syscall_interruption dsi;
+		boost::this_thread::disable_interruption di;
+		boost::this_thread::disable_syscall_interruption dsi;
 		P_DEBUG("Shutting down Phusion Passenger...");
+		LoggingKit::shutdown();
+		oxt::shutdown();
 		delete hooks;
 		hooks = NULL;
 	} catch (const thread_interrupted &) {
@@ -1565,6 +1628,15 @@ destroy_hooks(void *arg) {
 	return APR_SUCCESS;
 }
 
+static int
+preinit_module(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp) {
+	// When reloading Apache, global variables may be left at stale values,
+	// so here we reinitialize them.
+	hooks = NULL;
+	serverConfig = ServerConfig();
+	return OK;
+}
+
 static int
 init_module(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {
 	/*
@@ -1584,14 +1656,9 @@ init_module(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *
 	 * universal, i.e. it works for some people but not for others. So we got rid of the
 	 * hacks, and now we always initialize in the post_config hook.
 	 */
-	if (hooks == NULL) {
-		oxt::initialize();
-		SystemTime::initialize();
-	} else {
-		P_DEBUG("Restarting Phusion Passenger....");
-		delete hooks;
-		hooks = NULL;
-	}
+	oxt::initialize();
+	SystemTime::initialize();
+	LoggingKit::initialize();
 	try {
 		hooks = new Hooks(pconf, plog, ptemp, s);
 		apr_pool_cleanup_register(pconf, NULL,
@@ -1599,13 +1666,13 @@ init_module(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *
 			apr_pool_cleanup_null);
 		return OK;
 
-	} catch (const thread_interrupted &e) {
+	} catch (const boost::thread_interrupted &e) {
 		P_TRACE(2, "A system call was interrupted during mod_passenger "
 			"initialization. Apache might be restarting or shutting "
 			"down. Backtrace:\n" << e.backtrace());
 		return DECLINED;
 
-	} catch (const thread_resource_error &e) {
+	} catch (const boost::thread_resource_error &e) {
 		struct rlimit lim;
 		string pthread_threads_max;
 		int ret;
@@ -1668,13 +1735,13 @@ child_init(apr_pool_t *pchild, server_rec *s) {
 	}
 }
 
-#define DEFINE_REQUEST_HOOK(c_name, cpp_name)        \
-	static int c_name(request_rec *r) {          \
-		if (OXT_LIKELY(hooks != NULL)) {     \
-			return hooks->cpp_name(r);   \
-		} else {                             \
-			return DECLINED;             \
-		}                                    \
+#define DEFINE_REQUEST_HOOK(c_name, cpp_name) \
+	static int c_name(request_rec *r) { \
+		if (OXT_LIKELY(hooks != NULL)) { \
+			return hooks->cpp_name(r); \
+		} else { \
+			return DECLINED; \
+		} \
 	}
 
 DEFINE_REQUEST_HOOK(prepare_request_when_in_high_performance_mode, prepareRequestWhenInHighPerformanceMode)
@@ -1694,11 +1761,12 @@ DEFINE_REQUEST_HOOK(handle_request_when_not_in_high_performance_mode, handleRequ
  * Apache hook registration function.
  */
 void
-passenger_register_hooks(apr_pool_t *p) {
+registerHooks(apr_pool_t *p) {
 	static const char * const rewrite_module[] = { "mod_rewrite.c", NULL };
 	static const char * const dir_module[] = { "mod_dir.c", NULL };
 	static const char * const autoindex_module[] = { "mod_autoindex.c", NULL };
 
+	ap_hook_pre_config(preinit_module, NULL, NULL, APR_HOOK_MIDDLE);
 	ap_hook_post_config(init_module, NULL, NULL, APR_HOOK_MIDDLE);
 	ap_hook_child_init(child_init, NULL, NULL, APR_HOOK_MIDDLE);
 
@@ -1719,7 +1787,6 @@ passenger_register_hooks(apr_pool_t *p) {
 	ap_hook_handler(handle_request_when_not_in_high_performance_mode, NULL, NULL, APR_HOOK_LAST);
 }
 
-/**
- * @}
- */
 
+} // namespace Apache2Module
+} // namespace Passenger
