@@ -1,6 +1,6 @@
 # encoding: utf-8
 #  Phusion Passenger - https://www.phusionpassenger.com/
-#  Copyright (c) 2010-2016 Phusion Holding B.V.
+#  Copyright (c) 2010-2018 Phusion Holding B.V.
 #
 #  "Passenger", "Phusion Passenger" and "Union Station" are registered
 #  trademarks of Phusion Holding B.V.
@@ -30,14 +30,21 @@ module PhusionPassenger
   ###### Names and version numbers ######
 
   PACKAGE_NAME = 'passenger'
-  # Run 'rake src/cxx_supportlib/Constants.h' after changing this number.
-  VERSION_STRING = '5.0.30'
+  # Run 'rake src/cxx_supportlib/Constants.h configkit_schemas_inline_comments' after changing this number.
+  VERSION_STRING = '6.0.10'
 
-  PREFERRED_NGINX_VERSION = '1.10.1'
-  NGINX_SHA256_CHECKSUM = '1fd35846566485e03c0e318989561c135c598323ff349c503a6c14826487a801'
+  # Tip: find the SHA-256 with ./dev/nginx_version_sha2 <VERSION>
+  PREFERRED_NGINX_VERSION = '1.20.1'
+  NGINX_SHA256_CHECKSUM = 'e462e11533d5c30baa05df7652160ff5979591d291736cfa5edb9fd2edb48c49'
 
-  PREFERRED_PCRE_VERSION  = '8.39'
-  PCRE_SHA256_CHECKSUM = 'ccdf7e788769838f8285b3ee672ed573358202305ee361cfec7a4a4fb005bbc7'
+  # Packaging may be locked to an older version due to the specific module configuration being
+  # incompatible with the version we prefer (latest stable).
+  PACKAGING_PREFERRED_NGINX_VERSION = '1.20.1'
+  PACKAGING_NGINX_SHA256_CHECKSUM = 'e462e11533d5c30baa05df7652160ff5979591d291736cfa5edb9fd2edb48c49'
+
+  # sha256sum of the .tar.gz
+  PREFERRED_PCRE_VERSION  = '8.45'
+  PCRE_SHA256_CHECKSUM = '4e6ce03e0336e8b4a3d6c2b70b1c5e18590a5673a98186da90d4f33c23defc09'
 
   STANDALONE_INTERFACE_VERSION  = 1
 
@@ -168,7 +175,7 @@ module PhusionPassenger
 
   # Whether the current Phusion Passenger installation is installed
   # from a release package, e.g. an official gem or official tarball.
-  # Retruns false if e.g. the gem was built by the user, or if this
+  # Returns false if e.g. the gem was built by the user, or if this
   # install is from a git repository.
   def self.installed_from_release_package?
     File.exist?("#{resources_dir}/release.txt")
@@ -213,6 +220,7 @@ module PhusionPassenger
 
   def self.binaries_sites
     [
+      { :url => "https://github.com/phusion/passenger/releases/download/release-{{VERSION}}".freeze },
       { :url => "https://oss-binaries.phusionpassenger.com/binaries/passenger/by_release".freeze },
       { :url => "https://s3.amazonaws.com/phusion-passenger/binaries/passenger/by_release".freeze }
     ]
@@ -250,7 +258,7 @@ module PhusionPassenger
   end
 
 
-private
+  private
   def self.infer_install_spec
     filename = ENV['PASSENGER_LOCATION_CONFIGURATION_FILE']
     return filename if filename && !filename.empty?
@@ -320,7 +328,12 @@ private
     end
     if home.nil? || home.empty?
       require 'etc' if !defined?(Etc)
-      home = Etc.getpwuid(Process.uid).dir
+      begin
+        home = Etc.getpwuid(Process.uid).dir
+      rescue ArgumentError
+        # Unknown user.
+        home = ENV['HOME']
+      end
     end
     return home
   end
