@@ -1,5 +1,5 @@
 #  Phusion Passenger - https://www.phusionpassenger.com/
-#  Copyright (c) 2010-2013 Phusion Holding B.V.
+#  Copyright (c) 2010-2018 Phusion Holding B.V.
 #
 #  "Passenger", "Phusion Passenger" and "Union Station" are registered
 #  trademarks of Phusion Holding B.V.
@@ -144,7 +144,6 @@ module PhusionPassenger
         end
       end
 
-      flags << debugging_cflags
       flags << '-DHAS_ALLOCA_H' if has_alloca_h?
       flags << '-DHAVE_ACCEPT4' if has_accept4?
       flags << '-DHAS_SFENCE' if supports_sfence_instruction?
@@ -152,6 +151,7 @@ module PhusionPassenger
       flags << "-DPASSENGER_DEBUG -DBOOST_DISABLE_ASSERTS"
 
       if cc_or_cxx == :cxx
+        flags << debugging_cxxflags
         flags << cxx_11_flag if cxx_11_flag
 
         if cxx_supports_wno_unused_local_typedefs_flag?
@@ -159,11 +159,19 @@ module PhusionPassenger
           flags << "-Wno-unused-local-typedefs"
         end
 
+        if cxx_supports_wno_format_nonliteral_flag?
+          # SystemTools/UserDatabase.cpp uses snprintf() with a non-literal
+          # format string.
+          flags << '-Wno-format-nonliteral'
+        end
+
         # There are too many implementations of of the hash map!
         # Figure out the right one.
         check_unordered_map(flags, "std::unordered_map", "unordered_map", "HAS_UNORDERED_MAP") ||
           check_unordered_map(flags, "std::tr1::unordered_map", "unordered_map", "HAS_TR1_UNORDERED_MAP") ||
           check_hash_map(flags)
+      else
+        flags << debugging_cflags
       end
 
       if os_name_simple == "solaris"
@@ -195,6 +203,9 @@ module PhusionPassenger
         # http://groups.google.com/group/phusion-passenger/t/6b904a962ee28e5c
         # http://groups.google.com/group/phusion-passenger/browse_thread/thread/aad4bd9d8d200561
         flags << '-DBOOST_SP_USE_PTHREADS'
+      elsif linux_distro == :centos && os_version >= "8"
+        # _FORTIFY_SOURCE requires compiling with optimization (-O)
+        flags << '-O'
       end
 
       return flags.compact.map{ |str| str.strip }.join(" ").strip
